from typing import List, Optional, Union

import pandas as pd
from pandera.typing import DataFrame

from model_quality_report.experiment_base import ExperimentBase
from model_quality_report.model_comparison_report import (
    ModelComparisonReport,
    ModelComparisonReportType,
    ExperimentKeyType,
)
from model_quality_report.quality_report.base import QualityReportBase
from model_quality_report.schemas import (
    ModelQualityMetricsSchema,
    TrueAndPredictedFloatSchema,
    TrueAndPredictedBoolSchema,
)


class ReportAggregator:
    def __init__(self, experiments: List[ExperimentBase], n_jobs: Optional[int] = None) -> None:
        self._model_comparison = self._get_model_comparison(experiments=experiments, n_jobs=n_jobs)

    def run(self) -> List[ModelComparisonReportType]:
        return self._model_comparison.create_reports()

    def get_metrics(self, reports: List[ModelComparisonReportType]) -> DataFrame[ModelQualityMetricsSchema]:
        return self._model_comparison.get_metrics(reports=reports)

    def get_true_and_predicted_data(
        self, reports: List[ModelComparisonReportType]
    ) -> Union[DataFrame[TrueAndPredictedFloatSchema], DataFrame[TrueAndPredictedBoolSchema]]:
        return self._model_comparison.get_true_and_predicted_data(reports=reports).reset_index()

    def _get_model_comparison(
        self, experiments: List[ExperimentBase], n_jobs: Optional[int] = None
    ) -> ModelComparisonReport:
        return ModelComparisonReport(
            quality_reports=self._get_quality_reports(experiments=experiments),
            X_data_list=self._get_x_data_list(experiments=experiments),
            y_data_list=self._get_y_data_list(experiments=experiments),
            experiment_keys=self._get_experiment_keys(experiments=experiments),
            n_jobs=n_jobs,
        )

    @staticmethod
    def _get_quality_reports(experiments: List[ExperimentBase]) -> List[QualityReportBase]:
        return [experiment.get_quality_report() for experiment in experiments]

    @staticmethod
    def _get_x_data_list(experiments: List[ExperimentBase]) -> List[pd.DataFrame]:
        return [experiment.get_x_data() for experiment in experiments]

    @staticmethod
    def _get_y_data_list(experiments: List[ExperimentBase]) -> List[pd.Series]:
        return [experiment.get_y_data() for experiment in experiments]

    @staticmethod
    def _get_experiment_keys(experiments: List[ExperimentBase]) -> List[ExperimentKeyType]:
        return [experiment.get_experiment_key() for experiment in experiments]



from pandera import SchemaModel
from pandera.typing import Series, Float, String, Bool


class ModelQualityMetricsSchema(SchemaModel):
    metrics: Series[String]
    value: Series[Float]


class TrueAndPredictedFloatSchema(SchemaModel):
    true: Series[Float]
    predicted: Series[Float]


class TrueAndPredictedBoolSchema(SchemaModel):
    true: Series[Bool]
    predicted: Series[Bool]



from typing import List

import pandas as pd

from model_quality_report.splitters.base import SplitterBase, Splits, Split


class ByHorizon(SplitterBase):
    """
    Produces a list of splits of temporal data such that specific number of observations
    after the specified date is used as test data until the input data is exhausted.

    This class can be of use in Time Series analysis, where one wants to produce predictions for several steps ahead
    starting at different dates, in order to assess predictive power of a model by averaging errors across these dates
    for each specific maximum_horizon.

    The data used by this splitter has to have `pd.DatetimeIndex` as its index and it has to be sorted along this index.

    Example
    -------
    import pandas as pd
    from model_quality_report.splitters.temporal.cross_validation.by_horizon import ByHorizon

    lbl_date = 'date'
    dates = pd.Index(pd.date_range('2019-01-01', periods=6), name=lbl_date)
    X = pd.DataFrame({'a': [1, 2, 4, 5, 7, 20],
                      'b': [3, 5, 7, 10, 15, 30]},
                      index=dates)
    y = pd.Series([3, 6, 8, 10, 12, 30], index=dates)
    splitter = ByHorizon(start_split_date=dates[2], maximum_horizon=2)
    print(splitter.split(X, y))

    [(          a  b
    date
    2019-01-01  1  3
    2019-01-02  2  5,

                a   b
    date
    2019-01-03  4   7
    2019-01-04  5  10,

    date
    2019-01-01    3
    2019-01-02    6
    Freq: D, dtype: int64,

    date
    2019-01-03     8
    2019-01-04    10
    Freq: D, dtype: int64),

    (           a  b
    date
    2019-01-01  1  3
    2019-01-02  2  5
    2019-01-03  4  7,

                a   b
    date
    2019-01-04  5  10
    2019-01-05  7  15,

    date
    2019-01-01    3
    2019-01-02    6
    2019-01-03    8
    Freq: D, dtype: int64,

    date
    2019-01-04    10
    2019-01-05    12
    Freq: D, dtype: int64)]

    """

    lbl_date_column_name = "date_column_name"
    lbl_start_split_date = "start_split_date"
    lbl_split_dates = "split_dates"
    lbl_maximum_horizon = "maximum_horizon"

    def __init__(self, start_split_date: pd.Timestamp, maximum_horizon: int) -> None:
        self.start_split_date = start_split_date
        self.maximum_horizon = maximum_horizon

    def split(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        """
        Given df with features X, target series y and test config, data are split into test and training data
        taking data from last time_delta long period as test data and the rest as training data.

        :param X:
        :param y:
        :return: list of (X_train, X_test, y_train, y_test)
        """
        splits = list()
        for split_date in self._get_all_split_dates(X=X):
            date_filter = X.index <= self._get_latest_valid_date_for_test_set(X=X, split_date=split_date)
            splits.append(self._split_date_data(X=X[date_filter], y=y[date_filter], split_date=split_date))
        return splits

    @staticmethod
    def _split_date_data(X: pd.DataFrame, y: pd.Series, split_date: pd.Timestamp) -> Split:
        X_train = X.loc[X.index < split_date]
        X_test = X.loc[X.index >= split_date]
        y_train = y[X_train.index]
        y_test = y[X_test.index]
        return X_train, X_test, y_train, y_test

    def validate_parameters(self, X: pd.DataFrame, y: pd.Series) -> List:

        validation_error = list()

        if not isinstance(X.index, pd.DatetimeIndex):
            validation_error.append("X index is not DatetimeIndex.")

        if not isinstance(y.index, pd.DatetimeIndex):
            validation_error.append("y index is not DatetimeIndex.")

        try:
            if X.index.max() < self.start_split_date:
                validation_error.append(
                    "start split date {} is larger than max date {}".format(self.start_split_date, X.index.max())
                )
            if X.index.min() > self.start_split_date:
                validation_error.append(
                    "start split date {} is smaller than min date {}".format(self.start_split_date, X.index.min())
                )
        except TypeError as e:
            validation_error.append(str(e))

        if not pd.Index(X.index).is_monotonic:
            validation_error.append("data is not temporally sorted")

        try:
            if len(X.index > self.start_split_date) < self.maximum_horizon:
                validation_error.append(
                    "maximum_horizon is larger than the number of observations after start_split_date"
                )
        except TypeError as e:
            validation_error.append(str(e))

        return validation_error

    def get_parameters(self) -> dict:
        return {
            self.lbl_start_split_date: self.start_split_date,
            self.lbl_maximum_horizon: self.maximum_horizon,
        }

    def _get_all_split_dates(self, X: pd.DataFrame) -> list:
        return X.loc[self.start_split_date :].iloc[: -self.maximum_horizon].index.to_list()

    def _get_latest_valid_date_for_test_set(self, X: pd.DataFrame, split_date: pd.Timestamp) -> pd.Timestamp:
        return X.loc[split_date:].iloc[: self.maximum_horizon].index.max()



import os
import unittest

import pandas as pd

from model_quality_report.splitters.temporal.split_date import SplitDateDataSplitter


class TestSplitDateDataSplitter(unittest.TestCase):
    def setUp(self):
        __current_dir_path = os.path.dirname(os.path.abspath(__file__))
        test_data = pd.read_pickle(os.path.join(__current_dir_path, "../../data/date-value_df.pkl"))
        self.X = test_data.loc[:, ["date"]]
        self.X["date"] = pd.to_datetime(self.X["date"])
        self.y = test_data["value"]

    def test_splits_according_last_split_date(self):
        split_date = pd.Timestamp("2017-01-01")
        splitter = SplitDateDataSplitter(date_column_name="date", split_date=split_date)

        X_train, X_test, y_train, y_test = splitter.split(self.X, self.y)[0]

        self.assertGreaterEqual(X_test[splitter.date_column_name].min(), split_date)

        self.assertLessEqual(X_train[splitter.date_column_name].max(), split_date)

        self.assertSequenceEqual(list(X_train.index), list(y_train.index))
        self.assertSequenceEqual(list(X_test.index), list(y_test.index))

    def test_returns_error_if_split_date_too_large(self):
        split_date = pd.Timestamp("2100-01-01")
        splitter = SplitDateDataSplitter("date", split_date)
        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(str(split_date) in errors[0])

    def test_returns_error_if_split_date_too_small(self):
        split_date = pd.Timestamp("1900-01-01")
        splitter = SplitDateDataSplitter("date", split_date)
        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(str(split_date) in errors[0])

    def test_returns_error_if_wrong_timestamp_is_provided(self):
        splitter = SplitDateDataSplitter("date", "not_a_timestamp")
        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("Timestamp" in errors[0])

    def test_parameters_can_be_accessed_by_get_parameters(self):
        split_date = pd.Timestamp("1900-01-01")

        splitter = SplitDateDataSplitter("date", split_date)

        parameters = splitter.get_parameters()

        self.assertIsInstance(parameters, dict)
        self.assertDictEqual(
            {
                SplitDateDataSplitter.lbl_date_column_name: "date",
                SplitDateDataSplitter.lbl_split_date: split_date,
            },
            parameters,
        )

    def test_returns_proper_error_in_case_wrong_data_column_name_is_provided(self):
        not_existing_date_column = "not_existing_date_column"

        splitter = SplitDateDataSplitter(not_existing_date_column, pd.Timestamp("1900-01-01"))
        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("not_existing_date_column" in errors[0])



import logging
import time
from typing import List, TypedDict, Optional

import pandas as pd
from joblib import Parallel, delayed

from model_quality_report.quality_report.base import QualityReportBase, QualityReportType

ExperimentKeyType = dict


class ModelComparisonReportType(TypedDict):
    experiment_key: ExperimentKeyType
    report: QualityReportType
    errors: str


class ModelComparisonReport:
    """Model comparison report.

    It takes four lists of equal size: `QualityReportBase` instances, `X_data_list`, `y_data_list`,
    and flat dictionaries `experiment_keys`. The class now has only two public methods:
      - `create_reports` that simply returns the list quality reports for each experiment, and
      - `get_metrics` that extracts a single metrics DataFrame from multiple quality reports obtained above.
       The resulting DataFrame has experiment keys in corresponding columns.
    """

    lbl_metrics = QualityReportBase.lbl_metrics
    lbl_metric_value = QualityReportBase.lbl_metric_value

    lbl_experiment_key = "experiment_key"
    lbl_report = "report"
    lbl_errors = "errors"

    def __init__(
        self,
        quality_reports: List[QualityReportBase],
        X_data_list: List[pd.DataFrame],
        y_data_list: List[pd.Series],
        experiment_keys: List[ExperimentKeyType],
        n_jobs: Optional[int] = None,
    ) -> None:
        """

        :param quality_reports: list of `QualityReportBase` instances
        that would produce a quality report for each experiment.
        :param X_data_list: list of X data
        :param y_data_list: list of y data
        :param experiment_keys: list of flat dictionaries that provide a description for each experiment
        :param n_jobs: the maximum number of concurrently running jobs (see `joblib.Parallel`)
        """
        self._quality_reports = quality_reports
        self._X_data_list = X_data_list
        self._y_data_list = y_data_list
        self._experiment_keys = experiment_keys
        self._n_jobs = n_jobs
        self._logger = logging.getLogger(self.__class__.__name__)

    def create_reports(self) -> List[ModelComparisonReportType]:
        """
        Given a list of experiments compute quality reports for each and combine them in one dictionary.

        :return: dict containing the quality report
        """
        experiments = zip(self._experiment_keys, self._quality_reports, self._X_data_list, self._y_data_list)
        with Parallel(n_jobs=self._n_jobs, verbose=10, backend="loky") as parallel:
            results = parallel(
                delayed(create_reports_kernel)(experiment_key=experiment_key, quality_report=quality_report, X=X, y=y)
                for experiment_key, quality_report, X, y in experiments
            )
        return results

    def get_metrics(self, reports: List[ModelComparisonReportType]) -> pd.DataFrame:
        """Convert quality reports (only metrics part) into a single DataFrame.

        :return: DataFrame containing metric values for each experiment

        Example output
        --------------
                            metrics     value  model  exogenous
        0  explained_variance_score  0.949841  Model1      a, b
        1                      mape  0.177778  Model1      a, b
        2       mean_absolute_error  1.000000  Model1      a, b
        3        mean_squared_error  1.240000  Model1      a, b
        4     median_absolute_error  1.000000  Model1      a, b
        5                  r2_score  0.911429  Model1      a, b
        0  explained_variance_score  0.992172  Model2         a
        1                      mape  0.187302  Model2         a
        2       mean_absolute_error  0.952381  Model2         a
        3        mean_squared_error  1.142857  Model2         a
        4     median_absolute_error  1.142857  Model2         a
        5                  r2_score  0.992172  Model2         a
        0  explained_variance_score  0.973807  Model3         b
        1                      mape  0.158055  Model3         b
        2       mean_absolute_error  0.861451  Model3         b
        3        mean_squared_error  0.957465  Model3         b
        4     median_absolute_error  0.577017  Model3         b
        5                  r2_score  0.883552  Model3         b

        """
        results = list()
        for quality_report_instance, quality_report in zip(self._quality_reports, reports):
            report_df = quality_report_instance.get_metrics(report=quality_report[self.lbl_report]).assign(
                **quality_report[self.lbl_experiment_key]
            )
            results.append(report_df)
        return pd.concat(results)

    def get_true_and_predicted_data(self, reports: List[ModelComparisonReportType]) -> pd.DataFrame:
        """Convert quality reports (true and predicted values) into a single DataFrame.

        :param reports:
        :return: DataFrame containing metric values for each experiment

        Example output
        --------------
                            true  predicted    model y name
        date       horizon
        2019-01-03 0           8   9.600000  Model 1     y1
        2019-01-04 1          10  13.800000  Model 1     y1
        2019-01-05 2          12  21.000000  Model 1     y1
        2019-01-03 0           8   6.622642  Model 2     y2
        2019-01-04 1          10   0.226415  Model 2     y2
        2019-01-05 2          12   3.169811  Model 2     y2
        2019-01-04 0          10   7.916667  Model 2     y2

        """
        results = list()
        for quality_report_instance, quality_report in zip(self._quality_reports, reports):
            report_df = quality_report_instance.get_true_and_predicted_data(
                report=quality_report[self.lbl_report]
            ).assign(**quality_report[self.lbl_experiment_key])
            results.append(report_df)
        return pd.concat(results)


def create_reports_kernel(
    experiment_key: ExperimentKeyType,
    quality_report: QualityReportBase,
    X: pd.DataFrame,
    y: pd.Series,
) -> ModelComparisonReportType:
    logger = logging.getLogger(__name__)
    msg = (
        f"experiment key: {experiment_key}, "
        f"quality report: {quality_report.__class__.__name__}, "
        f"y data shape: {y.shape}, and name: {y.name}, "
        f"X data shape; {X.shape}, and columns: {X.columns.values}."
    )
    logger.info(f"Start quality report for {msg}")
    start_time = time.perf_counter()
    report = quality_report.create_quality_report(X=X, y=y)
    logger.info(f"Finish quality report for {msg}")
    logger.info(f"Total time: {time.perf_counter() - start_time} seconds.")
    return ModelComparisonReportType(
        experiment_key=experiment_key, report=report, errors=quality_report.get_errors().to_string()
    )



class QualityReportError:
    def __init__(self) -> None:
        self.error_list = list()

    def add(self, parameter: str) -> None:
        self.error_list.append(parameter)

    def is_empty(self) -> bool:
        if len(self.error_list) == 0:
            return True
        return False

    def to_string(self) -> str:
        return "\n".join(self.error_list)



from typing import List

import pandas as pd
from pandas.core.dtypes.common import is_datetime64_ns_dtype

from model_quality_report.splitters.base import SplitterBase, Splits, Split


class ByFrequency(SplitterBase):
    """
    Produces a list of splits of temporal data such that all observations
    after the specified date is used as test data until the input data is exhausted.

    This class can be of use in Time Series analysis, where one wants to produce predictions for several steps ahead
    starting at different dates, in order to assess predictive power of a model by averaging errors across these dates
    for each specific horizon.

    The X data used by this splitter has to have a `datetime64[ns]` column, and it has to be sorted along this column.
    The y data used by this splitter has to have `pd.DatetimeIndex` as its index and it has to be sorted along this index.

    Example
    -------
    import pandas as pd
    from model_quality_report.splitters.temporal.cross_validation.by_frequency import ByFrequency

    lbl_date = 'date'
    dates = pd.Index(pd.date_range('2019-01-01', periods=6), name=lbl_date)
    X = pd.DataFrame({'a': [1, 2, 4, 5, 7, 20],
                      'b': [3, 5, 7, 10, 15, 30],
                      lbl_date: dates})
    y = pd.Series([3, 6, 8, 10, 12, 30], index=dates)
    splitter = ByFrequency(start_split_date=dates[2], frequency="D", date_column_name=lbl_date)
    splitter.split(X, y)

    """

    lbl_date_column_name = "date_column_name"
    lbl_start_split_date = "start_split_date"
    lbl_split_dates = "split_dates"
    lbl_frequency = "frequency"

    def __init__(self, start_split_date: pd.Timestamp, date_column_name: str, frequency: str) -> None:
        self.start_split_date = start_split_date
        self.date_column_name = date_column_name
        self.frequency = frequency

    def split(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        """
        Given df with features X, target series y and test config, data are split into test and training data
        taking data from last time_delta long period as test data and the rest as training data.

        :param X:
        :param y:
        :return: list of (X_train, X_test, y_train, y_test)
        """
        return [self._split_date_data(X=X, y=y, split_date=split_date) for split_date in self._get_all_split_dates(X=X)]

    def _split_date_data(self, X: pd.DataFrame, y: pd.Series, split_date: pd.Timestamp) -> Split:
        X_train = X.loc[X[self.date_column_name] < split_date]
        X_test = X.loc[X[self.date_column_name] >= split_date]
        y_train = y.loc[y.index < split_date]
        y_test = y.loc[y.index >= split_date]
        return X_train, X_test, y_train, y_test

    def validate_parameters(self, X: pd.DataFrame, y: pd.Series) -> List:
        validation_error = list()

        if not isinstance(y.index, pd.DatetimeIndex):
            validation_error.append("y index is not DatetimeIndex.")

        if not is_datetime64_ns_dtype(X[self.date_column_name]):
            validation_error.append(f"{self.date_column_name} column is not datetime64[ns].")

        try:
            if X[self.date_column_name].max() < self.start_split_date:
                validation_error.append(
                    "start split date {} is larger than max date {}".format(
                        self.start_split_date, X[self.date_column_name].max()
                    )
                )
            if X[self.date_column_name].min() > self.start_split_date:
                validation_error.append(
                    "start split date {} is smaller than min date {}".format(
                        self.start_split_date, X[self.date_column_name].min()
                    )
                )
        except TypeError as e:
            validation_error.append(str(e))

        try:
            pd.date_range("2020-01-01", periods=1, freq=self.frequency)
        except ValueError as e:
            validation_error.append(str(e))

        if not pd.Index(X[self.date_column_name]).is_monotonic:
            validation_error.append("X data is not temporally sorted")

        if not pd.Index(y.index).is_monotonic:
            validation_error.append("y data is not temporally sorted")

        return validation_error

    def get_parameters(self) -> dict:
        return {
            self.lbl_start_split_date: self.start_split_date,
            self.lbl_date_column_name: self.date_column_name,
            self.lbl_frequency: self.frequency,
        }

    def _get_all_split_dates(self, X: pd.DataFrame) -> pd.DatetimeIndex:
        return pd.date_range(
            start=self.start_split_date,
            end=X[self.date_column_name].max(),
            freq=self.frequency,
        )



from typing import List

import pandas as pd

from model_quality_report.splitters.base import SplitterBase, Splits


class TimeDeltaDataSplitter(SplitterBase):
    """
    Splits data such that provided timedelta is used as test data
    """

    lbl_date_column_name = "date_column_name"
    lbl_time_delta = "time_delta"

    def __init__(self, date_column_name: str, time_delta: pd.Timedelta) -> None:
        self.date_column_name = date_column_name
        self.time_delta = time_delta

    def split(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        """
        Given df with features X, target series y, data are split into test and training data
        taking data from last time_delta long period as test data and the rest as training data.
        :param X:
        :param y:
        :return: X_train, X_test, y_train, y_test
        """
        split_date = X[self.date_column_name].max() - self.time_delta
        X_train = X[X[self.date_column_name] < split_date]
        X_test = X[X[self.date_column_name] >= split_date]
        y_train = y[X_train.index]
        y_test = y[X_test.index]
        return [(X_train, X_test, y_train, y_test)]

    def validate_parameters(self, X: pd.DataFrame, y: pd.Series) -> List:
        validation_error = list()

        try:
            X[self.date_column_name]
        except KeyError as e:
            validation_error.append("Column {} does not exist in DataFrame X".format(e))
            return validation_error

        try:
            if (X[self.date_column_name].max() - X[self.date_column_name].min()) < self.time_delta:
                validation_error.append("time_delta {} is too large for data".format(self.time_delta))
        except TypeError as e:
            validation_error.append(str(e))

        return validation_error

    def get_parameters(self):
        return {
            self.lbl_date_column_name: self.date_column_name,
            self.lbl_time_delta: self.time_delta,
        }



import copy
import logging
import warnings
from abc import ABC, abstractmethod
from typing import List, Union, Dict

import pandas as pd

from model_quality_report.model_base import ModelBase
from model_quality_report.quality_metrics import all_quality_metrics
from model_quality_report.quality_report.quality_report_error import QualityReportError
from model_quality_report.splitters.base import SplitterBase, Splits

PandasObj = Union[pd.Series, pd.DataFrame]
QualityReportType = Dict[str, PandasObj]


class QualityReportBase(ABC):
    """
    Base metaclass for _model quality reports.

    """

    lbl_metrics = "metrics"
    lbl_data = "data"
    lbl_true_values = "true"
    lbl_predicted_values = "predicted"
    lbl_metric_value = "value"

    def __init__(self, model: ModelBase, splitter: SplitterBase) -> None:
        self._model = copy.deepcopy(model)
        self._splitter = splitter
        self._errors = QualityReportError()
        self._logger = logging.getLogger(self.__class__.__name__)

    def create_quality_report(self, X: pd.DataFrame, y: pd.Series) -> QualityReportType:
        """
        Given a _model and a data set a report on _model performance is created.

        :return: dict containing the quality report
        """
        y_test_all = list()
        predictions_all = list()

        if not self._errors.is_empty():
            return dict()

        for X_train, X_test, y_train, y_test in self._split_data_for_quality_assessment(X=X, y=y):

            try:
                self._fit(X_train, y_train)
                predictions = self._predict(X_test)
            except (RuntimeError, ValueError) as e:
                self._logger.error(str(e))
                self._errors.add(str(e))
                return dict()
            else:
                y_test_all.append(y_test)
                predictions_all.append(predictions)

        return self._create_dict_from_test_values_and_predictions(y_trues=y_test_all, y_preds=predictions_all)

    def _split_data_for_quality_assessment(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        """
        Given features X and target y are split into training and test data.

        :return:
        """
        splits = list()
        validation_error = self._splitter.validate_parameters(X=X, y=y)
        if len(validation_error) == 0:
            splits = self._splitter.split(X=X, y=y)
        else:
            msg = "Split failed: {}".format(validation_error)
            self._logger.error(msg)
            self._errors.add(msg)
        return splits

    def _fit(self, X_train: pd.DataFrame, y_train: pd.Series) -> None:
        """
        Fits _model if it has a fit attribute

        :return:
        """
        try:
            self._model.fit(X_train, y_train)
        except (RuntimeError, AttributeError, ValueError) as e:
            self._logger.error(str(e))
            self._errors.add(str(e))

    def _predict(self, X_test: pd.DataFrame) -> pd.Series:
        """
        Predicts with _model if it has a predict attribute and returns predictions

        :return:
        """
        predictions = pd.Series(dtype=float)
        try:
            predictions = self._model.predict(X_test)
        except (RuntimeError, AttributeError, ValueError) as e:
            self._logger.error(str(e))
            self._errors.add(str(e))
        if not isinstance(predictions, pd.Series):
            predictions = pd.Series(predictions, index=X_test.index)
        return predictions

    @staticmethod
    def _true_and_predicted_values_as_pandas(y_true: PandasObj, y_pred: PandasObj) -> PandasObj:
        return pd.merge(y_true, y_pred, left_index=True, right_index=True)

    def _create_dict_from_test_values_and_predictions(
        self, y_trues: List[pd.Series], y_preds: List[pd.Series]
    ) -> QualityReportType:
        """Create report from test and predicted values of y.

        :param y_trues: list of true values of y
        :param y_preds: list of predicted values of y
        :return: quality report in dictionary format
        """
        y_trues = self._convert_list_of_data_to_pandas(data=y_trues, name=self.lbl_true_values)
        y_preds = self._convert_list_of_data_to_pandas(data=y_preds, name=self.lbl_predicted_values)

        if y_trues.index.names != y_preds.index.names:
            msg = (
                f"Test and predictions index names are different. "
                f"Test index names: {y_trues.index.names}. "
                f"Prediction index names: {y_preds.index.names}."
            )
            self._logger.error(msg)
            self._errors.add(msg)
            return dict()

        return {
            self.lbl_metrics: self._calculate_quality_metrics_as_pandas(y_true=y_trues, y_pred=y_preds),
            self.lbl_data: self._true_and_predicted_values_as_pandas(y_true=y_trues, y_pred=y_preds),
        }

    @staticmethod
    def _calculate_quality_metrics(y_true: pd.Series, y_pred: pd.Series) -> Dict[str, float]:
        """
        Given y_pred and y_true, various metrics are derived and returned as dict.

        :return: dict containing the quality metrics
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return {name: metric(y_true=y_true, y_pred=y_pred) for name, metric in all_quality_metrics.items()}

    @abstractmethod
    def _calculate_quality_metrics_as_pandas(self, y_true: PandasObj, y_pred: PandasObj) -> PandasObj:
        """
        Given y_pred and y_true, various metrics are derived and returned as pandas object.

        :return: Series or DataFrame containing the quality metrics
        """

    @abstractmethod
    def _convert_list_of_data_to_pandas(self, data: List[pd.Series], name: str = None) -> PandasObj:
        """Convert list of data to pandas object.

        This method is specific to quality report format.
        If a prediction is a scalar, then we return a Series of predictions for each split.
        If a prediction is array-like, e.g. time horizon along its dimension,
        then we return a DataFrame with rows for each split and columns for each horizon.

        :return: either Series or DataFrame depending on the report specifics
        """

    @classmethod
    def get_metrics(cls, report: dict) -> pd.DataFrame:
        """Convert quality report (only metrics part) from dictionary format into DataFrame.

        :return: DataFrame containing the quality report

        Example input
        -------------
        {'metrics':
        {'explained_variance_score': 0.9148351648351651,
        'mape': 0.4388888888888885,
        mean_absolute_error': 6.66666666666666,
        'mean_squared_error': 51.33333333333321,
        'median_absolute_error': 7.999999999999986,
        'r2_score': 0.365384615384617},
        'data':
        {'true': {4: 12, 3: 10, 5: 30},
        'predicted': {4: 20.999999999999993, 3: 12.999999999999998, 5: 37.999999999999986}}}

        Example output
        --------------
                            metrics      value
        0  explained_variance_score   0.914835
        1                      mape   0.438889
        2       mean_absolute_error   6.666667
        3        mean_squared_error  51.333333
        4     median_absolute_error   8.000000
        5                  r2_score   0.365385

        """
        return report.get(cls.lbl_metrics)

    @classmethod
    def get_true_and_predicted_data(cls, report: dict) -> pd.DataFrame:
        return report.get(cls.lbl_data)

    def get_errors(self) -> QualityReportError:
        return self._errors



import numpy as np
import pandas as pd
from sklearn.metrics import (
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    accuracy_score,
    precision_score,
    recall_score,
)


def _r2_score(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return r2_score(y_true=y_true, y_pred=y_pred)
    except (TypeError, ZeroDivisionError):
        return np.nan


def _explained_variance_score(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return explained_variance_score(y_true=y_true, y_pred=y_pred)
    except TypeError:
        return np.nan


def _mean_squared_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return mean_squared_error(y_true=y_true, y_pred=y_pred)
    except TypeError:
        return np.nan


def _mean_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.mean(y_true - y_pred))
    except TypeError:
        return np.nan


def _median_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.median(y_true - y_pred))
    except TypeError:
        return np.nan


def _mean_sign_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.mean(np.sign(y_true - y_pred)))
    except TypeError:
        return np.nan


def _mean_absolute_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return mean_absolute_error(y_true=y_true, y_pred=y_pred)
    except TypeError:
        return np.nan


def _median_absolute_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.median(np.abs(y_pred - y_true)))
    except TypeError:
        return np.nan


def _mean_percentage_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.mean((y_true - y_pred) / y_true))
    except (TypeError, ZeroDivisionError):
        return np.nan


def _median_percentage_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.median((y_true - y_pred) / y_true))
    except (TypeError, ZeroDivisionError):
        return np.nan


def _mean_absolute_percentage_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.mean(np.abs((y_true - y_pred) / y_true)))
    except (TypeError, ZeroDivisionError):
        return np.nan


def _median_absolute_percentage_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.median(np.abs((y_true - y_pred) / y_true)))
    except (TypeError, ZeroDivisionError):
        return np.nan


def _mean_absolute_cum_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return mean_absolute_error(y_true=np.cumsum(y_true), y_pred=np.cumsum(y_pred))
    except TypeError:
        return np.nan


def _median_absolute_cum_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.median(np.abs(np.cumsum(y_pred) - np.cumsum(y_true))))
    except TypeError:
        return np.nan


def _mean_absolute_percentage_cum_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.mean(np.abs((np.cumsum(y_true) - np.cumsum(y_pred)) / np.cumsum(y_true))))
    except (TypeError, ZeroDivisionError):
        return np.nan


def _median_absolute_percentage_cum_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.median(np.abs((np.cumsum(y_true) - np.cumsum(y_pred)) / np.cumsum(y_true))))
    except (TypeError, ZeroDivisionError):
        return np.nan


def _mean_sign_cum_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return float(np.mean(np.sign(np.cumsum(y_true) - np.cumsum(y_pred))))
    except TypeError:
        return np.nan


def _accuracy(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return accuracy_score(y_true=y_true, y_pred=y_pred, normalize=True)
    except ValueError:
        return np.nan


def _precision(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return precision_score(y_true=y_true, y_pred=y_pred, zero_division=0)
    except ValueError:
        return np.nan


def _recall(y_true: pd.Series, y_pred: pd.Series) -> float:
    try:
        return recall_score(y_true=y_true, y_pred=y_pred, zero_division=0)
    except ValueError:
        return np.nan


regression_quality_metrics = {
    "r2_score": _r2_score,
    "explained_variance_score": _explained_variance_score,
    "mean_squared_error": _mean_squared_error,
    "mean_error": _mean_error,
    "median_error": _median_error,
    "mean_sign_error": _mean_sign_error,
    "mean_absolute_error": _mean_absolute_error,
    "median_absolute_error": _median_absolute_error,
    "mean_percentage_error": _mean_percentage_error,
    "median_percentage_error": _median_percentage_error,
    "mean_absolute_percentage_error": _mean_absolute_percentage_error,
    "median_absolute_percentage_error": _median_absolute_percentage_error,
    "mean_absolute_cum_error": _mean_absolute_cum_error,
    "median_absolute_cum_error": _median_absolute_cum_error,
    "mean_absolute_percentage_cum_error": _mean_absolute_percentage_cum_error,
    "median_absolute_percentage_cum_error": _median_absolute_percentage_cum_error,
    "mean_sign_cum_error": _mean_sign_cum_error,
}

classification_quality_metrics = {"accuracy": _accuracy, "precision": _precision, "recall": _recall}


all_quality_metrics = {**regression_quality_metrics, **classification_quality_metrics}



import unittest

from model_quality_report.quality_report.quality_report_error import QualityReportError


class TestQualityReportError(unittest.TestCase):
    def test_error_object_is_initially_empty(self):
        error = QualityReportError()
        self.assertTrue(error.is_empty())

    def test_is_empty_returns_false_after_adding_errors(self):
        error = QualityReportError()

        error.add("This does not work")
        self.assertFalse(error.is_empty())

        error.add("This also does not work")
        self.assertFalse(error.is_empty())
        self.assertTrue(len(error.error_list) == 2)

    def test_to_string_method_returns_string(self):
        error = QualityReportError()

        error.add("This does not work")
        self.assertFalse(error.is_empty())

        error.add("This also does not work")
        self.assertTrue(isinstance(error.to_string(), str))



import pandas as pd
from sklearn.model_selection import train_test_split

from model_quality_report.splitters.base import SplitterBase, Splits


class RandomDataSplitter(SplitterBase):
    """
    Splits data randomly using the sklearn model_selection.train_test_split
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def split(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        return [train_test_split(X, y, **self.kwargs)]

    def validate_parameters(self, X: pd.DataFrame, y: pd.Series) -> list:
        validation_error = list()
        try:
            train_test_split(X, y, **self.kwargs)
        except Exception as e:
            validation_error.append(str(e))
        return validation_error

    def get_parameters(self):
        return self.kwargs



from __future__ import annotations

from abc import ABC, abstractmethod

import pandas as pd


class ModelBase(ABC):
    """
    Base metaclass for model provided to quality report library. It is basically used to provide type hints.

    """

    @abstractmethod
    def fit(self, X_train: pd.DataFrame, y_train: pd.Series) -> ModelBase:
        """
        Each model passed to quality report must have a fit function
        """

    @abstractmethod
    def predict(self, X_test: pd.DataFrame) -> pd.Series:
        """
        Each model passed to quality report must have a predict function
        """



import unittest

import pandas as pd
from pandas.testing import assert_frame_equal, assert_series_equal

from model_quality_report.splitters.random import RandomDataSplitter


class TestRandomDataSplitter(unittest.TestCase):
    def test_data_are_split_randomly(self):

        X = pd.DataFrame({"a": [1, 2, 3, 4, 5], "b": ["a", "b", "c", "d", "e"]})
        y = pd.Series(data=range(5))

        splitter = RandomDataSplitter(test_size=0.33, random_state=2)
        X_train, X_test, y_train, y_test = splitter.split(X, y)[0]

        assert_frame_equal(
            X_train,
            pd.DataFrame({"a": [2, 4, 1], "b": ["b", "d", "a"]}, index=[1, 3, 0]),
        )
        assert_frame_equal(X_test, pd.DataFrame({"a": [3, 5], "b": ["c", "e"]}, index=[2, 4]))
        assert_series_equal(y_train, pd.Series([1, 3, 0], index=[1, 3, 0]))
        assert_series_equal(y_test, pd.Series([2, 4], index=[2, 4]))

    def test_error_is_returned_if_test_size_is_too_large(self):
        X = pd.DataFrame({"a": [1, 2, 3, 4, 5], "b": ["a", "b", "c", "d", "e"]})
        y = pd.Series(data=range(5))

        splitter = RandomDataSplitter(test_size=100, random_state=2)

        errors = splitter.validate_parameters(X, y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("test_size=100" in errors[0])

    def test_parameters_can_be_accessed_by_get_parameters(self):
        splitter = RandomDataSplitter(test_size=100, random_state=2)

        parameters = splitter.get_parameters()

        self.assertIsInstance(parameters, dict)
        self.assertDictEqual({"test_size": 100, "random_state": 2}, parameters)



from typing import List, Type

import pandas as pd

from model_quality_report.experiment_base import ExperimentBase
from model_quality_report.model_base import ModelBase
from model_quality_report.model_comparison_report import ExperimentKeyType
from model_quality_report.quality_report.base import QualityReportBase
from model_quality_report.quality_report.crossvalidation_timeseries import CrossValidationTimeSeriesQualityReport
from model_quality_report.splitters.base import SplitterBase
from model_quality_report.splitters.temporal.cross_validation.by_horizon import ByHorizon
from tests.models_for_tests import LinearModelWrapper, ClassifierModelWrapper


class LinearModelExperiment(ExperimentBase):
    def __init__(self, x_data: pd.DataFrame, y_data: pd.Series, exog_cols: List[str], experiment_key: str) -> None:
        self._x_data = x_data
        self._y_data = y_data
        self._exog_cols = exog_cols
        self._experiment_key = experiment_key

    @property
    def _quality_report_class(self) -> Type[QualityReportBase]:
        return CrossValidationTimeSeriesQualityReport

    def get_experiment_key(self) -> ExperimentKeyType:
        return {self._experiment_key: str(self._exog_cols)}

    def get_x_data(self) -> pd.DataFrame:
        return self._x_data

    def get_y_data(self) -> pd.Series:
        return self._y_data

    def _get_model(self) -> ModelBase:
        return LinearModelWrapper(exog_cols=self._exog_cols)

    def _get_splitter(self) -> SplitterBase:
        return ByHorizon(start_split_date=self._x_data.index[2], maximum_horizon=3)


class ClassifierModelExperiment(LinearModelExperiment):
    def _get_model(self) -> ModelBase:
        return ClassifierModelWrapper(exog_cols=self._exog_cols)



import unittest
from typing import Optional

import pandas as pd
from pandera.errors import SchemaErrors
from parameterized import parameterized

from model_quality_report.report_aggregator import ReportAggregator
from model_quality_report.schemas import (
    ModelQualityMetricsSchema,
    TrueAndPredictedFloatSchema,
    TrueAndPredictedBoolSchema,
)
from tests.linear_model_experiment import LinearModelExperiment, ClassifierModelExperiment


class TestReportAggregator(unittest.TestCase):
    def setUp(self):
        dates = pd.date_range("2019-01-01", periods=8)
        index = pd.DatetimeIndex(dates, name="date")
        self.X = pd.DataFrame(
            {
                "a": [1.1, 2, 4, 5, 7, 20, 6, 11],
                "b": [3.3, 5, 7, 10, 15, 30, 22, 48],
                "c": [8.8, 1, 0, 17, 11, 20, 2, 4],
            },
            index=index,
        )
        self.y_data_float = pd.Series([3.4, 6, 8, 10, 12, 30, 23, 5], name="y1", index=index)
        self.y_data_bool = pd.Series([True, False, True, True, False, False, False, True], name="y2", index=index)

    @parameterized.expand([(None,), (-1,)])
    def test_report_aggregator_with_linear_model(self, n_jobs: Optional[int] = None):
        experiments = list()
        experiment_keys = ["key1", "key2"]
        for exog_cols, experiment_key in zip((["a", "b"], ["b", "c"]), experiment_keys):
            experiments.append(
                LinearModelExperiment(
                    x_data=self.X, y_data=self.y_data_float, exog_cols=exog_cols, experiment_key=experiment_key
                )
            )

        aggregator = ReportAggregator(experiments=experiments, n_jobs=n_jobs)
        reports = aggregator.run()
        metrics = aggregator.get_metrics(reports=reports)
        true_and_predicted = aggregator.get_true_and_predicted_data(reports=reports)

        ModelQualityMetricsSchema.validate(metrics, lazy=True)
        TrueAndPredictedFloatSchema.validate(true_and_predicted, lazy=True)
        with self.assertRaises(SchemaErrors):
            TrueAndPredictedBoolSchema.validate(true_and_predicted, lazy=True)
        for key in experiment_keys:
            self.assertIn(key, metrics.columns)
            self.assertIn(key, true_and_predicted.columns)

    @parameterized.expand([(None,), (-1,)])
    def test_report_aggregator_with_classifier_model(self, n_jobs: Optional[int] = None):
        experiments = list()
        experiment_keys = ["key1", "key2"]
        for exog_cols, experiment_key in zip((["a", "b"], ["b", "c"]), experiment_keys):
            experiments.append(
                ClassifierModelExperiment(
                    x_data=self.X, y_data=self.y_data_bool, exog_cols=exog_cols, experiment_key=experiment_key
                )
            )

        aggregator = ReportAggregator(experiments=experiments, n_jobs=n_jobs)
        reports = aggregator.run()
        metrics = aggregator.get_metrics(reports=reports)
        true_and_predicted = aggregator.get_true_and_predicted_data(reports=reports)

        ModelQualityMetricsSchema.validate(metrics, lazy=True)
        TrueAndPredictedBoolSchema.validate(true_and_predicted, lazy=True)
        with self.assertRaises(SchemaErrors):
            TrueAndPredictedFloatSchema.validate(true_and_predicted, lazy=True)
        for key in experiment_keys:
            self.assertIn(key, metrics.columns)
            self.assertIn(key, true_and_predicted.columns)



from typing import List

import numpy as np
import pandas as pd

from model_quality_report.quality_report.base import QualityReportBase


class CrossValidationTimeSeriesQualityReport(QualityReportBase):
    """Cross-validation time series quality report.

    Collects quality metrics when model predictions are in general non-scalar, e.g. for several time steps ahead.

    """

    lbl_horizon = "horizon"

    def _calculate_quality_metrics_as_pandas(self, y_true: pd.DataFrame, y_pred: pd.DataFrame) -> pd.DataFrame:
        return (
            y_true.join(y_pred)
            .dropna()
            .infer_objects()
            .groupby(self.lbl_horizon)
            .apply(
                lambda df_true_pred: pd.Series(
                    self._calculate_quality_metrics(
                        y_true=df_true_pred[self.lbl_true_values],
                        y_pred=df_true_pred[self.lbl_predicted_values],
                    )
                )
            )
            .rename_axis(self.lbl_metrics, axis=1)
            .rename_axis(self.lbl_horizon, axis=0)
            .stack()
            .rename(self.lbl_metric_value)
            .reset_index()
        )

    def _convert_list_of_data_to_pandas(self, data: List[pd.Series], name: str = None) -> pd.DataFrame:
        df_list = [pd.DataFrame({name: s, self.lbl_horizon: np.arange(s.shape[0])}) for s in data]
        return pd.concat(df_list).set_index(self.lbl_horizon, append=True)



from typing import List

import pandas as pd

from model_quality_report.quality_report.base import QualityReportBase


class RegressionQualityReport(QualityReportBase):
    def _calculate_quality_metrics_as_pandas(self, y_true: pd.Series, y_pred: pd.Series) -> pd.Series:
        return (
            pd.Series(self._calculate_quality_metrics(y_true=y_true, y_pred=y_pred))
            .rename(self.lbl_metric_value)
            .rename_axis(self.lbl_metrics, axis=0)
            .reset_index()
        )

    def _convert_list_of_data_to_pandas(self, data: List[pd.Series], name: str = None) -> pd.Series:
        return pd.Series(data[0], name=name)



import unittest

import numpy as np
import pandas as pd
from sklearn.metrics import (
    explained_variance_score,
    mean_squared_error,
    mean_absolute_error,
    r2_score,
    accuracy_score,
    precision_score,
    recall_score,
)

from model_quality_report.quality_metrics import regression_quality_metrics, classification_quality_metrics


class TestQualityMetrics(unittest.TestCase):
    def test_quality_metrics(self):
        y_true = pd.Series([3, -0.5, 2, 7])
        y_pred = pd.Series([2.5, 0.0, 2, 8])
        rqm = regression_quality_metrics

        self.assertEqual(rqm.get("r2_score")(y_true, y_pred), r2_score(y_true, y_pred))
        self.assertEqual(
            rqm.get("explained_variance_score")(y_true, y_pred),
            explained_variance_score(y_true, y_pred),
        )
        self.assertEqual(rqm.get("mean_squared_error")(y_true, y_pred), mean_squared_error(y_true, y_pred))
        self.assertEqual(rqm.get("mean_absolute_error")(y_true, y_pred), mean_absolute_error(y_true, y_pred))
        self.assertEqual(rqm.get("median_absolute_error")(y_true, y_pred), np.median(np.abs(y_pred - y_true)))
        self.assertEqual(rqm.get("mean_error")(y_true, y_pred), np.mean(y_true - y_pred))
        self.assertEqual(rqm.get("median_error")(y_true, y_pred), np.median(y_true - y_pred))
        self.assertEqual(rqm.get("mean_percentage_error")(y_true, y_pred), np.mean((y_true - y_pred) / y_true))
        self.assertEqual(
            rqm.get("median_percentage_error")(y_true, y_pred),
            np.median((y_true - y_pred) / y_true),
        )
        self.assertEqual(
            rqm.get("mean_absolute_percentage_error")(y_true, y_pred),
            np.mean(np.abs((y_true - y_pred) / y_true)),
        )
        self.assertEqual(
            rqm.get("median_absolute_percentage_error")(y_true, y_pred),
            np.median(np.abs((y_true - y_pred) / y_true)),
        )
        self.assertEqual(
            rqm.get("mean_absolute_cum_error")(y_true, y_pred),
            mean_absolute_error(np.cumsum(y_true), np.cumsum(y_pred)),
        )
        self.assertEqual(
            rqm.get("median_absolute_cum_error")(y_true, y_pred),
            np.median(np.abs(np.cumsum(y_pred) - np.cumsum(y_true))),
        )
        self.assertEqual(
            rqm.get("mean_absolute_percentage_cum_error")(y_true, y_pred),
            np.mean(np.abs((np.cumsum(y_true) - np.cumsum(y_pred)) / np.cumsum(y_true))),
        )
        self.assertEqual(
            rqm.get("median_absolute_percentage_cum_error")(y_true, y_pred),
            np.median(np.abs((np.cumsum(y_true) - np.cumsum(y_pred)) / np.cumsum(y_true))),
        )
        self.assertEqual(
            rqm.get("mean_sign_cum_error")(y_true, y_pred),
            np.mean(np.sign(np.cumsum(y_true) - np.cumsum(y_pred))),
        )

        cqm = classification_quality_metrics

        self.assertTrue(np.isnan(cqm.get("accuracy")(y_true, y_pred)))
        self.assertTrue(np.isnan(cqm.get("precision")(y_true, y_pred)))
        self.assertTrue(np.isnan(cqm.get("recall")(y_true, y_pred)))

    def test_quality_metrics_with_bool_data(self):
        y_true = pd.Series([True, True, False])
        y_pred = pd.Series([True, False, False])
        rqm = regression_quality_metrics

        self.assertTrue(np.isnan(rqm.get("r2_score")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("explained_variance_score")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_squared_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_absolute_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_absolute_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_absolute_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_absolute_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_percentage_error")(y_true, y_pred)))
        self.assertFalse(np.isnan(rqm.get("mean_absolute_cum_error")(y_true, y_pred)))
        self.assertFalse(np.isnan(rqm.get("median_absolute_cum_error")(y_true, y_pred)))
        self.assertFalse(np.isnan(rqm.get("mean_absolute_percentage_cum_error")(y_true, y_pred)))
        self.assertFalse(np.isnan(rqm.get("median_absolute_percentage_cum_error")(y_true, y_pred)))
        self.assertFalse(np.isnan(rqm.get("mean_sign_cum_error")(y_true, y_pred)))

        cqm = classification_quality_metrics

        self.assertAlmostEqual(cqm.get("accuracy")(y_true, y_pred), 0.6666, places=3)
        self.assertEqual(cqm.get("precision")(y_true, y_pred), 1)
        self.assertEqual(cqm.get("recall")(y_true, y_pred), 0.5)

    def test_r2_score(self):
        y_true = pd.Series([3])
        y_pred = pd.Series([2.5])
        self.assertTrue(np.isnan(regression_quality_metrics.get("r2_score")(y_true, y_pred)))

    def test_classification_metrics(self):
        cqm = classification_quality_metrics

        y_true = pd.Series([True])
        y_pred = pd.Series([True])
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), 1)
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), accuracy_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("precision")(y_true, y_pred), 1)
        self.assertEqual(cqm.get("precision")(y_true, y_pred), precision_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("recall")(y_true, y_pred), 1)
        self.assertEqual(cqm.get("recall")(y_true, y_pred), recall_score(y_true=y_true, y_pred=y_pred))

        y_true = pd.Series([False])
        y_pred = pd.Series([False])
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), 1)
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), accuracy_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("precision")(y_true, y_pred), 0)
        self.assertEqual(
            cqm.get("precision")(y_true, y_pred), precision_score(y_true=y_true, y_pred=y_pred, zero_division=0)
        )
        self.assertEqual(cqm.get("recall")(y_true, y_pred), 0)
        self.assertEqual(cqm.get("recall")(y_true, y_pred), recall_score(y_true=y_true, y_pred=y_pred, zero_division=0))

        y_true = pd.Series([1, 0])
        y_pred = pd.Series([0, 1])
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), 0)
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), accuracy_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("precision")(y_true, y_pred), 0)
        self.assertEqual(cqm.get("precision")(y_true, y_pred), precision_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("recall")(y_true, y_pred), 0)
        self.assertEqual(cqm.get("recall")(y_true, y_pred), recall_score(y_true=y_true, y_pred=y_pred))

        y_true = pd.Series([1, 1])
        y_pred = pd.Series([0, 1])
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), 0.5)
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), accuracy_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("precision")(y_true, y_pred), 1)
        self.assertEqual(cqm.get("precision")(y_true, y_pred), precision_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("recall")(y_true, y_pred), 0.5)
        self.assertEqual(cqm.get("recall")(y_true, y_pred), recall_score(y_true=y_true, y_pred=y_pred))

        y_true = pd.Series([0, 0])
        y_pred = pd.Series([0, 1])
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), 0.5)
        self.assertEqual(cqm.get("accuracy")(y_true, y_pred), accuracy_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("precision")(y_true, y_pred), 0)
        self.assertEqual(cqm.get("precision")(y_true, y_pred), precision_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(cqm.get("recall")(y_true, y_pred), 0)
        self.assertEqual(cqm.get("recall")(y_true, y_pred), recall_score(y_true=y_true, y_pred=y_pred, zero_division=0))

    def test_division_by_zero(self):
        y_true = pd.Series([0, 0, 0])
        y_pred = pd.Series([0, 0, 0])
        rqm = regression_quality_metrics

        self.assertTrue(np.isnan(rqm.get("mean_absolute_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_absolute_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_absolute_percentage_cum_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_absolute_percentage_cum_error")(y_true, y_pred)))

        y_true = pd.Series([False, False, False])
        y_pred = pd.Series([False, False, False])
        rqm = regression_quality_metrics

        self.assertTrue(np.isnan(rqm.get("mean_absolute_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_absolute_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_percentage_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("mean_absolute_percentage_cum_error")(y_true, y_pred)))
        self.assertTrue(np.isnan(rqm.get("median_absolute_percentage_cum_error")(y_true, y_pred)))



import unittest

import numpy as np
import pandas as pd
from parameterized import parameterized_class

from model_quality_report.quality_metrics import all_quality_metrics
from model_quality_report.quality_report.crossvalidation_timeseries import (
    CrossValidationTimeSeriesQualityReport,
)
from model_quality_report.splitters.random import RandomDataSplitter
from model_quality_report.splitters.temporal.cross_validation.by_frequency import (
    ByFrequency,
)
from model_quality_report.splitters.temporal.cross_validation.by_horizon import (
    ByHorizon,
)
from tests.models_for_tests import (
    LinearModelWrapper,
    LinearModelWrapperWithIndexName,
)

by_horizon = ByHorizon(start_split_date=pd.Timestamp("2019-01-03"), maximum_horizon=2)


by_frequency = ByFrequency(start_split_date=pd.Timestamp("2019-01-03"), frequency="D", date_column_name="date")


@parameterized_class(
    [
        {"splitter": by_horizon, "len_metrics": 2},
        {"splitter": by_frequency, "len_metrics": 5},
    ]
)
class Test2DQualityReport(unittest.TestCase):
    def setUp(self):
        self.lbl_date = "date"
        dates = pd.date_range("2019-01-01", periods=7)
        self.index = pd.DatetimeIndex(dates, name=self.lbl_date)
        self.X = pd.DataFrame(
            {
                "a": [1, 2, 4, 5, 7, 20, 6],
                "b": [3, 5, 7, 10, 15, 30, 22],
                "date": dates,
            },
            index=self.index,
        )
        self.y = pd.Series([3, 6, 8, 10, 12, 30, 23], index=self.index)
        self.model = LinearModelWrapper(exog_cols=["a", "b"])
        self.maximum_horizon = 2
        self.frequency = "D"

    def test_calculate_metrics(self):
        y_true = pd.Series([3, -0.5, 2, 7])
        y_pred = pd.Series([2.5, 0.0, 2, 8])

        metrics = CrossValidationTimeSeriesQualityReport._calculate_quality_metrics(y_true, y_pred)
        result_dict = {name: metric(y_true, y_pred) for name, metric in all_quality_metrics.items()}

        self.assertTrue(isinstance(metrics, dict))
        self.assertDictEqual(metrics, result_dict)

    def test_splitting_fails_and_returns_error(self):
        wrong_splitter = RandomDataSplitter(test_size=10)
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, wrong_splitter)
        quality_report._split_data_for_quality_assessment(self.X, self.y)

        self.assertTrue("Split failed" in quality_report._errors.to_string())

    def test_fit_does_not_create_error_when_proper_model_is_provided(self):
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X, self.y
        )[0]
        quality_report._fit(X_train, y_train)

        self.assertTrue(quality_report._errors.is_empty())

    def test_quality_report_contains_error_if_fit_attribute_is_not_present(self):
        quality_report = CrossValidationTimeSeriesQualityReport("no_model_but_string", self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X,
            self.y,
        )[0]
        quality_report._fit(X_train, y_train)

        self.assertFalse(quality_report._errors.is_empty())
        self.assertTrue("fit" in quality_report._errors.to_string())

    def test_predict_does_not_create_error_when_proper_model_is_provided(self):
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X,
            self.y,
        )[0]
        quality_report._fit(X_train, y_train)
        prediction_result = quality_report._predict(X_test)

        self.assertTrue(quality_report._errors.is_empty())
        self.assertIsInstance(prediction_result, pd.Series)
        self.assertTrue(len(prediction_result), len(y_train))

    def test_error_when_test_predict_not_aligned(self):
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, self.splitter)

        def _convert_list_of_data_to_pandas(data, name):
            df_list = [pd.DataFrame({name: df, quality_report.lbl_horizon: np.arange(df.shape[0])}) for df in data]
            return (
                pd.concat(df_list)
                .set_index(quality_report.lbl_horizon, append=True)
                .pipe(lambda x: x.sort_values(by=x.columns[0]))
            )

        quality_report._convert_list_of_data_to_pandas = _convert_list_of_data_to_pandas

        result = quality_report.create_quality_report(self.X, self.y)

        self.assertTrue(quality_report._errors.is_empty())
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(
            [
                CrossValidationTimeSeriesQualityReport.lbl_metrics,
                CrossValidationTimeSeriesQualityReport.lbl_data,
            ],
            list(result.keys()),
        )
        self.assertTrue(isinstance(result[CrossValidationTimeSeriesQualityReport.lbl_metrics], pd.DataFrame))
        self.assertTrue(
            isinstance(
                result.get(CrossValidationTimeSeriesQualityReport.lbl_data),
                pd.DataFrame,
            )
        )

    def test_quality_report_contains_error_if_fit_and_predict_attribute_are_not_present(
        self,
    ):
        for model in [None, "no_model_but_string"]:
            quality_report = CrossValidationTimeSeriesQualityReport(model, self.splitter)
            (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
                self.X,
                self.y,
            )[0]
            quality_report._fit(X_train, y_train)
            prediction_result = quality_report._predict(X_test)

            self.assertFalse(quality_report._errors.is_empty())
            self.assertTrue("fit" in quality_report._errors.to_string())
            self.assertTrue("predict" in quality_report._errors.to_string())
            self.assertIsInstance(prediction_result, pd.Series)

    def test_quality_report_is_properly_returned(self):
        index_name = "another"
        self.model = LinearModelWrapperWithIndexName(exog_cols=["a", "b"], index_name=index_name)
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, self.splitter)

        result = quality_report.create_quality_report(self.X, self.y)

        self.assertEqual(result, dict())
        self.assertFalse(quality_report.get_errors().is_empty())
        self.assertIn(index_name, quality_report.get_errors().to_string())

    def test_quality_report_is_empty_if_index_names_are_different(self):
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, self.splitter)

        result = quality_report.create_quality_report(self.X, self.y)

        self.assertTrue(isinstance(result, dict))
        self.assertEqual(
            [
                CrossValidationTimeSeriesQualityReport.lbl_metrics,
                CrossValidationTimeSeriesQualityReport.lbl_data,
            ],
            list(result.keys()),
        )
        self.assertTrue(isinstance(result[CrossValidationTimeSeriesQualityReport.lbl_metrics], pd.DataFrame))
        self.assertTrue(
            isinstance(
                result.get(CrossValidationTimeSeriesQualityReport.lbl_data),
                pd.DataFrame,
            )
        )

    def test_create_quality_report_data_format(self):
        quality_report = CrossValidationTimeSeriesQualityReport(self.model, self.splitter)
        result = quality_report.create_quality_report(X=self.X, y=self.y)

        metrics = quality_report.get_metrics(report=result)
        self.assertIsInstance(metrics, pd.DataFrame)
        self.assertEqual(
            set(metrics.columns),
            {
                quality_report.lbl_metrics,
                quality_report.lbl_metric_value,
                quality_report.lbl_horizon,
            },
        )
        pd.testing.assert_frame_equal(metrics, result.get(quality_report.lbl_metrics))

        data = quality_report.get_true_and_predicted_data(report=result)
        self.assertIsInstance(data, pd.DataFrame)
        self.assertEqual(
            set(data.columns),
            {quality_report.lbl_true_values, quality_report.lbl_predicted_values},
        )
        self.assertEqual({quality_report.lbl_horizon, self.lbl_date}, set(data.index.names))
        pd.testing.assert_frame_equal(data, result.get(quality_report.lbl_data))



from abc import ABC, abstractmethod
from typing import List, Tuple

import pandas as pd

Split = Tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]
Splits = List[Split]


class SplitterBase(ABC):
    """
    Base metaclass for splitting data.
    """

    @abstractmethod
    def split(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        """
        Given a data frame containing features and a series containing corresponding targets, a split into training
        and test data is performed, which are subsequently returned.
        :return: X_train, X_test, y_train, y_test
        """

    @abstractmethod
    def validate_parameters(self, X: pd.DataFrame, y: pd.Series) -> List[str]:
        """
        A splitting strategy can depend on various parameters, like e.g. the size of the test set or a date  after which
        all data are declared as test data. This functions validates that data and parameters fit to each other.
        :return: error
        """

    @abstractmethod
    def get_parameters(self) -> dict:
        """
        Returns parameters of Splitter as dict
        :return: dict of parameters
        """



import os
import unittest

import pandas as pd

from model_quality_report.splitters.temporal.cross_validation.by_horizon import (
    ByHorizon,
)


class TestByHorizon(unittest.TestCase):
    def setUp(self):
        __current_dir_path = os.path.dirname(os.path.abspath(__file__))
        test_data = (
            pd.read_pickle(os.path.join(__current_dir_path, "../../../data/date-value_df.pkl"))
            .drop_duplicates(subset=["date"])
            .assign(date=lambda x: pd.to_datetime(x["date"]))
            .set_index("date")
            .sort_index()
        )
        self.X = test_data.assign(date=lambda x: x.index)
        self.y = test_data["value"]

    def test_correct_number_of_splits(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        splits = splitter.split(self.X, self.y)

        self.assertEqual(
            len(splits),
            self.X[self.X["date"] >= start_split_date].shape[0] - maximum_horizon,
        )

    def test_split_date_boundaries(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        splits = splitter.split(self.X, self.y)

        for X_train, X_test, y_train, y_test in splits:
            self.assertGreaterEqual(X_test.index.min(), start_split_date)
            self.assertLess(X_train.index.max(), X_test.index.min())

    def test_equality_of_split_x_an_y_indices(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        splits = splitter.split(self.X, self.y)

        for X_train, X_test, y_train, y_test in splits:
            self.assertSequenceEqual(list(X_train.index), list(y_train.index))
            self.assertSequenceEqual(list(X_test.index), list(y_test.index))

    def test_length_of_test_data(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        splits = splitter.split(self.X, self.y)

        for X_train, X_test, y_train, y_test in splits:
            self.assertEqual(X_test.shape[0], maximum_horizon)
            self.assertEqual(y_test.shape[0], maximum_horizon)

    def test_increment_of_train_data(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        splits = splitter.split(self.X, self.y)

        for split, split_next in zip(splits[:-1], splits[1:]):
            X_train, _, y_train, _ = split
            X_train_next, _, y_train_next, _ = split_next
            self.assertEqual(X_train.shape[0], X_train_next.shape[0] - 1)
            self.assertEqual(y_train.shape[0], y_train_next.shape[0] - 1)

    def test_returns_error_if_no_datetime_index(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)
        errors = splitter.validate_parameters(self.X.reset_index(drop=True), self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("DatetimeIndex" in errors[0])

        errors = splitter.validate_parameters(self.X, self.y.reset_index(drop=True))

        self.assertTrue(len(errors) > 0)
        self.assertTrue("DatetimeIndex" in errors[0])

    def test_returns_error_if_split_date_too_large(self):
        start_split_date = pd.Timestamp("2100-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(str(start_split_date) in errors[0])

    def test_returns_error_if_split_date_too_small(self):
        start_split_date = pd.Timestamp("1900-01-01")
        maximum_horizon = 10
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(str(start_split_date) in errors[0])

    def test_returns_error_if_maximum_horizon_too_large(self):
        start_split_date = pd.Timestamp("2017-01-01")
        maximum_horizon = 1000
        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("maximum_horizon" in errors[0])

    def test_returns_error_if_wrong_timestamp_is_provided(self):
        splitter = ByHorizon(start_split_date="not_a_timestamp", maximum_horizon=10)
        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("Timestamp" in errors[0])

    def test_parameters_can_be_accessed_by_get_parameters(self):
        start_split_date = pd.Timestamp("1900-01-01")
        maximum_horizon = 10

        splitter = ByHorizon(start_split_date=start_split_date, maximum_horizon=maximum_horizon)

        parameters = splitter.get_parameters()

        self.assertIsInstance(parameters, dict)
        self.assertDictEqual(
            {
                ByHorizon.lbl_start_split_date: start_split_date,
                ByHorizon.lbl_maximum_horizon: maximum_horizon,
            },
            parameters,
        )



import unittest

import pandas as pd
from sklearn import linear_model

from model_quality_report.quality_metrics import all_quality_metrics
from model_quality_report.quality_report.regression import (
    RegressionQualityReport,
)
from model_quality_report.splitters.random import RandomDataSplitter


class TestRegressionQualityReport(unittest.TestCase):
    def setUp(self):
        self.X = pd.DataFrame({"a": [1, 2, 4, 5, 7, 20], "b": [3, 5, 7, 10, 15, 30]})
        self.y = pd.Series([3, 6, 8, 10, 12, 30])
        self.model = linear_model.LinearRegression()
        self.test_size = 0.5
        self.splitter = RandomDataSplitter(test_size=self.test_size)

    def test_calculate_metrics(self):
        y_true = pd.Series([3, -0.5, 2, 7])
        y_pred = pd.Series([2.5, 0.0, 2, 8])

        metrics = RegressionQualityReport._calculate_quality_metrics(y_true, y_pred)
        result_dict = {name: metric(y_true, y_pred) for name, metric in all_quality_metrics.items()}

        self.assertTrue(isinstance(metrics, dict))
        self.assertDictEqual(metrics, result_dict)

    def test_splitting_returns_training_and_test_data(self):
        quality_report = RegressionQualityReport(self.model, self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X, self.y
        )[0]
        self.assertIsNotNone(X_train)
        self.assertIsNotNone(X_test)
        self.assertIsNotNone(y_train)
        self.assertIsNotNone(y_test)

        self.assertAlmostEqual(X_train.shape[0] / self.X.shape[0], 1 - self.test_size)
        self.assertAlmostEqual(y_train.shape[0] / self.X.shape[0], 1 - self.test_size)
        self.assertListEqual(list(X_train.index), list(y_train.index))

        self.assertAlmostEqual(X_test.shape[0] / self.X.shape[0], self.test_size)
        self.assertAlmostEqual(y_test.shape[0] / self.X.shape[0], self.test_size)
        self.assertListEqual(list(X_test.index), list(y_test.index))

    def test_splitting_fails_and_returns_error(self):
        wrong_splitter = RandomDataSplitter(test_size=10)
        quality_report = RegressionQualityReport(self.model, wrong_splitter)
        quality_report._split_data_for_quality_assessment(self.X, self.y)

        self.assertTrue("Split failed" in quality_report._errors.to_string())

    def test_fit_does_not_create_error_when_proper_model_is_provided(self):
        quality_report = RegressionQualityReport(self.model, self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X, self.y
        )[0]
        quality_report._fit(X_train, y_train)

        self.assertTrue(quality_report._errors.is_empty())

    def test_quality_report_contains_error_if_fit_attribute_is_not_present(self):
        quality_report = RegressionQualityReport("no_model_but_string", self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X,
            self.y,
        )[0]
        quality_report._fit(X_train, y_train)

        self.assertFalse(quality_report._errors.is_empty())
        self.assertTrue("fit" in quality_report._errors.to_string())

    def test_predict_does_not_create_error_when_proper_model_is_provided(self):
        quality_report = RegressionQualityReport(self.model, self.splitter)
        (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
            self.X,
            self.y,
        )[0]
        quality_report._fit(X_train, y_train)
        prediction_result = quality_report._predict(X_test)

        self.assertTrue(quality_report._errors.is_empty())
        self.assertIsInstance(prediction_result, pd.Series)
        self.assertTrue(len(prediction_result), len(y_train))

    def test_quality_report_contains_error_if_fit_and_predict_attribute_are_not_present(
        self,
    ):
        for model in [None, "no_model_but_string"]:
            quality_report = RegressionQualityReport(model, self.splitter)
            (X_train, X_test, y_train, y_test,) = quality_report._split_data_for_quality_assessment(
                self.X,
                self.y,
            )[0]
            quality_report._fit(X_train, y_train)
            prediction_result = quality_report._predict(X_test)

            self.assertFalse(quality_report._errors.is_empty())
            self.assertTrue("fit" in quality_report._errors.to_string())
            self.assertTrue("predict" in quality_report._errors.to_string())
            self.assertIsInstance(prediction_result, pd.Series)

    def test_quality_report_is_properly_returned(self):
        quality_report = RegressionQualityReport(self.model, self.splitter)

        result = quality_report.create_quality_report(self.X, self.y)

        self.assertTrue(isinstance(result, dict))
        self.assertEqual(
            [quality_report.lbl_metrics, quality_report.lbl_data],
            list(result.keys()),
        )
        metrics = quality_report.get_metrics(report=result)
        self.assertTrue(isinstance(metrics, pd.DataFrame))
        self.assertEqual(
            set(metrics.columns),
            {
                quality_report.lbl_metrics,
                quality_report.lbl_metric_value,
            },
        )
        pd.testing.assert_frame_equal(metrics, result.get(quality_report.lbl_metrics))

        data = quality_report.get_true_and_predicted_data(report=result)
        self.assertTrue(isinstance(data, pd.DataFrame))
        self.assertEqual(
            set(data.columns),
            {quality_report.lbl_true_values, quality_report.lbl_predicted_values},
        )
        pd.testing.assert_frame_equal(data, result.get(quality_report.lbl_data))

    def test_quality_report_metrics_format(self):
        quality_report = RegressionQualityReport(self.model, self.splitter)
        result = quality_report.create_quality_report(X=self.X, y=self.y)
        report_df = RegressionQualityReport.get_metrics(report=result)

        self.assertIsInstance(report_df, pd.DataFrame)
        self.assertEqual(
            set(report_df.columns),
            {quality_report.lbl_metrics, quality_report.lbl_metric_value},
        )



import os
import random
import unittest

import pandas as pd

from model_quality_report.splitters.sorted import SortedDataSplitter


class TestTemporalDataSplitter(unittest.TestCase):
    def setUp(self):
        __current_dir_path = os.path.dirname(os.path.abspath(__file__))
        self.test_data = pd.read_pickle(os.path.join(__current_dir_path, "../data/date-value_df.pkl"))

    def test_date_column_is_properly_split(self):
        date_column_name = "date"
        value_column_name = "value"

        X = self.test_data.loc[:, [date_column_name]]
        X[date_column_name] = pd.to_datetime(X[date_column_name])
        y = self.test_data[value_column_name]

        test_size = 0.2
        for ascending in [True, False]:
            splitter = SortedDataSplitter(
                sortable_column_name=date_column_name,
                test_size=test_size,
                ascending=ascending,
            )

            X_train, X_test, y_train, y_test = splitter.split(X, y)[0]

            self.assertAlmostEqual(X_test.shape[0] / X.shape[0], test_size)
            self.assertAlmostEqual(X_train.shape[0] / X.shape[0], 1 - test_size)

            self.assertAlmostEqual(y_test.shape[0] / y.shape[0], test_size)
            self.assertAlmostEqual(y_train.shape[0] / y.shape[0], 1 - test_size)

            if ascending:
                self.assertTrue(X_train[date_column_name].max() <= X_test[date_column_name].min())
            else:
                self.assertTrue(X_train[date_column_name].max() >= X_test[date_column_name].min())

    def test_numeric_column_is_properly_split(self):
        sortable_column = "sortable_column"
        random.seed(2)
        X = pd.DataFrame(
            data=[random.randint(0, 1000) for i in range(1000)],
            columns=[sortable_column],
        )
        y = pd.Series(range(1000))

        test_size = 0.5
        splitter = SortedDataSplitter(sortable_column_name=sortable_column, test_size=test_size, ascending=True)

        X_train, X_test, y_train, y_test = splitter.split(X, y)[0]

        self.assertAlmostEqual(X_test.shape[0] / X.shape[0], test_size)
        self.assertAlmostEqual(X_train.shape[0] / X.shape[0], 1 - test_size)

        self.assertAlmostEqual(y_test.shape[0] / y.shape[0], test_size)
        self.assertAlmostEqual(y_train.shape[0] / y.shape[0], 1 - test_size)

        self.assertTrue(X_train[sortable_column].max() <= X_test[sortable_column].min())

    def test_validate_parameters_returns_error_if_column_does_not_exist(self):
        not_existing_column = "not_existing_column"

        random.seed(2)
        X = pd.DataFrame(
            data=[random.randint(0, 1000) for i in range(1000)],
            columns=["sortable_column"],
        )
        y = pd.Series(range(1000))

        splitter = SortedDataSplitter(sortable_column_name=not_existing_column, test_size=0.5, ascending=True)
        errors = splitter.validate_parameters(X, y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(not_existing_column in errors[0])

    def test_validate_parameters_returns_error_if_column_can_not_be_sorted(self):
        sortable_column = "sortable_column"

        random.seed(2)
        X = pd.DataFrame(
            data=[random.randint(0, 1000) for i in range(1000)],
            columns=[sortable_column],
        )
        X.loc[0, sortable_column] = "no_number"
        y = pd.Series(range(1000))

        splitter = SortedDataSplitter(sortable_column_name=sortable_column, test_size=0.5, ascending=True)
        errors = splitter.validate_parameters(X, y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("<" in errors[0])

    def test_validate_parameters_returns_error_if_test_size_is_out_of_bounds(self):
        sortable_column = "sortable_column"

        random.seed(2)
        X = pd.DataFrame(
            data=[random.randint(0, 1000) for i in range(1000)],
            columns=[sortable_column],
        )
        y = pd.Series(range(1000))

        for test_size in [-1, 10]:
            splitter = SortedDataSplitter(
                sortable_column_name=sortable_column,
                test_size=test_size,
                ascending=True,
            )
            errors = splitter.validate_parameters(X, y)

            self.assertTrue(len(errors) > 0)
            self.assertTrue("test_size" in errors[0])

    def test_parameters_can_be_accessed_by_get_parameters(self):
        date_column_name = "date"
        test_size = 0.2
        ascending = True

        splitter = SortedDataSplitter(
            sortable_column_name=date_column_name,
            test_size=test_size,
            ascending=ascending,
        )

        parameters = splitter.get_parameters()

        self.assertIsInstance(parameters, dict)
        self.assertDictEqual(
            {
                SortedDataSplitter.lbl_sortable_column_name: date_column_name,
                SortedDataSplitter.lbl_test_size: test_size,
                SortedDataSplitter.lbl_ascending: ascending,
            },
            parameters,
        )



from abc import abstractmethod, ABC
from typing import Type

import pandas as pd

from model_quality_report.model_base import ModelBase
from model_quality_report.model_comparison_report import ExperimentKeyType
from model_quality_report.quality_report.base import QualityReportBase
from model_quality_report.splitters.base import SplitterBase


class ExperimentBase(ABC):
    def get_quality_report(self) -> QualityReportBase:
        return self._quality_report_class(model=self._get_model(), splitter=self._get_splitter())

    @abstractmethod
    def get_experiment_key(self) -> ExperimentKeyType:
        pass

    @abstractmethod
    def get_x_data(self) -> pd.DataFrame:
        pass

    @abstractmethod
    def get_y_data(self) -> pd.Series:
        pass

    @property
    @abstractmethod
    def _quality_report_class(self) -> Type[QualityReportBase]:
        pass

    @abstractmethod
    def _get_model(self) -> ModelBase:
        pass

    @abstractmethod
    def _get_splitter(self) -> SplitterBase:
        pass



import unittest

import pandas as pd

from tests.linear_model_experiment import LinearModelExperiment


class TestCrossValidationModelComparisonReport(unittest.TestCase):
    def setUp(self):
        dates = pd.date_range("2019-01-01", periods=8)
        index = pd.DatetimeIndex(dates, name="date")
        self.X = pd.DataFrame(
            {
                "a": [1, 2, 4, 5, 7, 20, 6, 11],
                "b": [3, 5, 7, 10, 15, 30, 22, 48],
                "c": [8, 1, 0, 17, 11, 20, 2, 4],
            },
            index=index,
        )
        self.y_data = pd.Series([3, 6, 8, 10, 12, 30, 23, 5], name="y1", index=index)

    def test_experiment_init(self):
        exog_cols = ["a", "b"]
        experiment_key = "key1"
        experiment = LinearModelExperiment(
            x_data=self.X, y_data=self.y_data, exog_cols=exog_cols, experiment_key=experiment_key
        )
        self.assertEqual(experiment.get_experiment_key(), {experiment_key: str(exog_cols)})
        pd.testing.assert_frame_equal(experiment.get_x_data(), self.X)
        pd.testing.assert_series_equal(experiment.get_y_data(), self.y_data)



from typing import List

import pandas as pd

from model_quality_report.splitters.base import SplitterBase, Splits


class SortedDataSplitter(SplitterBase):
    """
    Splits data after sorting along a given column such that last test_size data are used as test_data. Thereby,
    test_size must be larger than 0 and smaller than 1.
    """

    lbl_sortable_column_name = "sortable_column_name"
    lbl_test_size = "test_size"
    lbl_ascending = "ascending"

    def __init__(self, sortable_column_name: str, test_size: float, ascending: bool = True) -> None:
        self.sortable_column_name = sortable_column_name
        self.test_size = test_size
        self.ascending = ascending

    def split(self, X: pd.DataFrame, y: pd.Series) -> Splits:
        """
        Given df with features X, target series y data are split into test and training data
        such that
        :param X:
        :param y:
        :return: X_train, X_test, y_train, y_test
        """
        X = X.sort_values(by=self.sortable_column_name, ascending=self.ascending)
        X_train = X.iloc[: int((1 - self.test_size) * X.shape[0])]
        X_test = X.iloc[int((1 - self.test_size) * X.shape[0]) :]
        y_train = y[X_train.index]
        y_test = y[X_test.index]
        return [(X_train, X_test, y_train, y_test)]

    def validate_parameters(self, X: pd.DataFrame, y: pd.Series) -> List:
        validation_error = list()

        try:
            X[self.sortable_column_name]
        except KeyError as e:
            validation_error.append("Column {} does not exist in DataFrame X".format(e))
            return validation_error

        try:
            X.sort_values(by=self.sortable_column_name)
        except TypeError as e:
            validation_error.append(str(e))

        if (self.test_size < 0) | (self.test_size > 1):
            validation_error.append("test_size should be between 0 and 1")

        return validation_error

    def get_parameters(self) -> dict:
        return {
            self.lbl_sortable_column_name: self.sortable_column_name,
            self.lbl_test_size: self.test_size,
            self.lbl_ascending: self.ascending,
        }



import os
import unittest

import pandas as pd
from numpy.random import randint

from model_quality_report.splitters.temporal.cross_validation.by_frequency import (
    ByFrequency,
)


class TestByFrequency(unittest.TestCase):
    def setUp(self):
        __current_dir_path = os.path.dirname(os.path.abspath(__file__))
        test_data = (
            pd.read_pickle(os.path.join(__current_dir_path, "../../../data/date-value_df.pkl"))
            .drop_duplicates(subset=["date"])
            .assign(date=lambda x: pd.to_datetime(x["date"]))
            .sort_values(by="date")
            .reset_index(drop=True)
        )
        self.X = test_data.assign(**{"exog": lambda x: randint(100, size=test_data.shape[0])}).drop(columns=["value"])
        self.y = test_data.set_index("date")["value"]

    def test_correct_number_of_splits(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        splits = splitter.split(self.X, self.y)

        self.assertEqual(len(splits), 47)

    def test_split_date_boundaries(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        splits = splitter.split(self.X, self.y)

        for X_train, X_test, y_train, y_test in splits:
            self.assertGreaterEqual(X_test["date"].min(), start_split_date)
            self.assertLess(X_train["date"].max(), X_test["date"].min())
            self.assertGreaterEqual(y_test.index.min(), start_split_date)
            self.assertLess(y_train.index.max(), y_test.index.min())

    def test_increment_of_train_data(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        splits = splitter.split(self.X, self.y)

        for split, split_next in zip(splits[:-1], splits[1:]):
            X_train, _, y_train, _ = split
            X_train_next, _, y_train_next, _ = split_next
            self.assertLessEqual(X_train.shape[0], X_train_next.shape[0])
            self.assertLessEqual(y_train.shape[0], y_train_next.shape[0])

    def test_returns_error_if_no_datetime_index(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        errors = splitter.validate_parameters(self.X, self.y.reset_index(drop=True))

        self.assertTrue(len(errors) > 0)
        self.assertTrue("DatetimeIndex" in errors[0])

    def test_returns_error_if_split_date_too_large(self):
        start_split_date = pd.Timestamp("2100-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(str(start_split_date) in errors[0])

    def test_returns_error_if_split_date_too_small(self):
        start_split_date = pd.Timestamp("1900-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue(str(start_split_date) in errors[0])

    def test_returns_error_if_wrong_frequency_is_provided(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "wrong_freq"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("wrong_freq" in errors[0])

    def test_returns_error_if_wrong_date_col_is_provided(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "wrong_freq"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="exog",
        )

        errors = splitter.validate_parameters(self.X, self.y)

        self.assertTrue(len(errors) > 0)
        self.assertTrue("exog" in errors[0])

    def test_parameters_can_be_accessed_by_get_parameters(self):
        start_split_date = pd.Timestamp("2017-01-01")
        frequency = "W-FRI"
        splitter = ByFrequency(
            start_split_date=start_split_date,
            frequency=frequency,
            date_column_name="date",
        )

        parameters = splitter.get_parameters()

        self.assertIsInstance(parameters, dict)
        self.assertDictEqual(
            {
                ByFrequency.lbl_start_split_date: start_split_date,
                ByFrequency.lbl_frequency: frequency,
                ByFrequency.lbl_date_column_name: "date",
            },
            parameters,
        )



import unittest

import pandas as pd
from parameterized import parameterized_class
from sklearn import linear_model

from model_quality_report.model_comparison_report import ModelComparisonReport
from model_quality_report.quality_report.base import QualityReportBase
from model_quality_report.quality_report.crossvalidation_timeseries import (
    CrossValidationTimeSeriesQualityReport,
)
from model_quality_report.quality_report.regression import (
    RegressionQualityReport,
)
from model_quality_report.splitters.random import RandomDataSplitter
from model_quality_report.splitters.temporal.cross_validation.by_horizon import (
    ByHorizon,
)
from tests.models_for_tests import LinearModelWrapper


@parameterized_class(("n_jobs",), [(None,), (-1,)])
class TestCrossValidationModelComparisonReport(unittest.TestCase):
    def setUp(self):
        self.lbl_date = "date"
        self.lbl_model = "model"
        self.lbl_yname = "y name"
        self.lbl_model_name1 = "Model 1"
        self.lbl_model_name2 = "Model 2"
        dates = pd.date_range("2019-01-01", periods=8)
        self.index = pd.DatetimeIndex(dates, name=self.lbl_date)
        self.X = pd.DataFrame(
            {
                "a": [1, 2, 4, 5, 7, 20, 6, 11],
                "b": [3, 5, 7, 10, 15, 30, 22, 48],
                "c": [8, 1, 0, 17, 11, 20, 2, 4],
            },
            index=self.index,
        )
        y1 = pd.Series([3, 6, 8, 10, 12, 30, 23, 5], name="y1", index=self.index)
        y2 = pd.Series([3, 6, 8, 10, 12, 30, 23, 5], name="y2", index=self.index)
        model1 = LinearModelWrapper(exog_cols=["a", "b"])
        model2 = LinearModelWrapper(exog_cols=["b", "c"])
        self.maximum_horizon = 3
        self.splitter = ByHorizon(start_split_date=dates[2], maximum_horizon=self.maximum_horizon)
        self.quality_reports = [
            CrossValidationTimeSeriesQualityReport(model=model, splitter=self.splitter)
            for model in [model1, model2, model2]
        ]
        self.X_data_list = [self.X for _ in range(3)]
        self.y_data_list = [y1, y2, y1]
        self.experiment_keys = [
            {self.lbl_model: self.lbl_model_name1, self.lbl_yname: y1.name},
            {self.lbl_model: self.lbl_model_name2, self.lbl_yname: y2.name},
            {self.lbl_model: self.lbl_model_name2, self.lbl_yname: y1.name},
        ]

    def test_basic_model_comparison_results(self):
        model_comparison = ModelComparisonReport(
            quality_reports=self.quality_reports,
            X_data_list=self.X_data_list,
            y_data_list=self.y_data_list,
            experiment_keys=self.experiment_keys,
            n_jobs=self.n_jobs,
        )
        reports = model_comparison.create_reports()

        self.assertIsInstance(reports, list)
        for report in reports:
            self.assertIsInstance(report, dict)
            self.assertEqual(len(report), 3)

        for report, experiment_key in zip(reports, self.experiment_keys):
            self.assertEqual(report.get(model_comparison.lbl_experiment_key), experiment_key)
            self.assertIsInstance(report.get(model_comparison.lbl_report), dict)
            self.assertIsInstance(report.get(model_comparison.lbl_errors), str)

    def test_report_conversion(self):
        model_comparison = ModelComparisonReport(
            quality_reports=self.quality_reports,
            X_data_list=self.X_data_list,
            y_data_list=self.y_data_list,
            experiment_keys=self.experiment_keys,
            n_jobs=self.n_jobs,
        )
        reports = model_comparison.create_reports()
        result_df = model_comparison.get_metrics(reports=reports)

        self.assertIsInstance(result_df, pd.DataFrame)
        self.assertIn(self.lbl_model, result_df.columns)
        self.assertIn(self.lbl_yname, result_df.columns)
        self.assertIn(model_comparison.lbl_metrics, result_df.columns)
        self.assertIn(model_comparison.lbl_metric_value, result_df.columns)

        result_df = model_comparison.get_true_and_predicted_data(reports=reports)

        self.assertIsInstance(result_df, pd.DataFrame)
        self.assertIn(QualityReportBase.lbl_true_values, result_df.columns)
        self.assertIn(QualityReportBase.lbl_predicted_values, result_df.columns)
        self.assertIn(self.lbl_model, result_df.columns)
        self.assertIn(self.lbl_yname, result_df.columns)
        self.assertIn(CrossValidationTimeSeriesQualityReport.lbl_horizon, result_df.index.names)
        self.assertIn(self.lbl_date, result_df.index.names)


@parameterized_class(("n_jobs",), [(None,), (-1,)])
class TestRegressionModelComparisonReport(unittest.TestCase):
    def setUp(self):
        self.X = pd.DataFrame({"a": [1, 2, 4, 5, 7, 20], "b": [3, 5, 7, 10, 15, 30]})
        self.y = pd.Series([3, 6, 8, 10, 12, 30], name="y")
        self.model = linear_model.LinearRegression()
        self.test_size = 0.5
        self.splitter = RandomDataSplitter(test_size=self.test_size)
        self.lbl_model = "model"
        self.lbl_exog = "exogenous"
        self.lbl_model_name = "Model"
        self.exog_cols1 = ["a", "b"]
        self.exog_cols2 = ["a"]
        self.exog_cols3 = ["b"]

        self.quality_reports = [RegressionQualityReport(model=self.model, splitter=self.splitter) for _ in range(3)]
        self.X_data_list = [self.X[cols] for cols in [self.exog_cols1, self.exog_cols2, self.exog_cols3]]
        self.y_data_list = [self.y for _ in range(3)]
        self.experiment_keys = [
            {
                self.lbl_model: self.lbl_model_name,
                self.lbl_exog: ", ".join(cols),
            }
            for cols in [self.exog_cols1, self.exog_cols2, self.exog_cols3]
        ]

    def test_basic_model_comparison_results(self):
        model_comparison = ModelComparisonReport(
            quality_reports=self.quality_reports,
            X_data_list=self.X_data_list,
            y_data_list=self.y_data_list,
            experiment_keys=self.experiment_keys,
            n_jobs=self.n_jobs,
        )
        if self.n_jobs is None:
            with self.assertLogs(level="INFO") as caplog:
                reports = model_comparison.create_reports()
        else:
            reports = model_comparison.create_reports()

        self.assertIsInstance(reports, list)
        for report in reports:
            self.assertIsInstance(report, dict)
            self.assertEqual(len(report), 3)

        for report, experiment_key, quality_report_instance, y_data, x_data in zip(
            reports, self.experiment_keys, self.quality_reports, self.y_data_list, self.X_data_list
        ):
            self.assertEqual(report[model_comparison.lbl_experiment_key], experiment_key)
            self.assertIsInstance(report[model_comparison.lbl_report], dict)
            self.assertIsInstance(report[model_comparison.lbl_errors], str)
            if self.n_jobs is None:
                self.assertIn(str(experiment_key), "".join(caplog.output))
                self.assertIn(str(quality_report_instance.__class__.__name__), "".join(caplog.output))
                self.assertIn(y_data.name, "".join(caplog.output))
                self.assertIn(str(y_data.shape), "".join(caplog.output))
                self.assertIn(str(x_data.columns.values), "".join(caplog.output))
                self.assertIn(str(x_data.shape), "".join(caplog.output))

    def test_report_conversion(self):
        model_comparison = ModelComparisonReport(
            quality_reports=self.quality_reports,
            X_data_list=self.X_data_list,
            y_data_list=self.y_data_list,
            experiment_keys=self.experiment_keys,
            n_jobs=self.n_jobs,
        )
        reports = model_comparison.create_reports()
        result_df = model_comparison.get_metrics(reports=reports)

        self.assertIsInstance(result_df, pd.DataFrame)
        self.assertIn(self.lbl_model, result_df.columns)
        self.assertIn(self.lbl_exog, result_df.columns)
        self.assertIn(model_comparison.lbl_metrics, result_df.columns)
        self.assertIn(model_comparison.lbl_metric_value, result_df.columns)

        result_df = model_comparison.get_true_and_predicted_data(reports=reports)

        self.assertIsInstance(result_df, pd.DataFrame)
        self.assertIn(QualityReportBase.lbl_true_values, result_df.columns)
        self.assertIn(QualityReportBase.lbl_predicted_values, result_df.columns)
        self.assertIn(self.lbl_model, result_df.columns)
        self.assertIn(self.lbl_exog, result_df.columns)



from __future__ import annotations

from typing import List

import pandas as pd
from numpy import ndarray
from sklearn.linear_model import LinearRegression
from sklearn.svm import SVC

from model_quality_report.model_base import ModelBase


class LinearModelWrapper(ModelBase):
    def __init__(self, exog_cols: List[str]) -> None:
        self._exog_cols = exog_cols
        self.model = LinearRegression()

    def fit(self, X_train: pd.DataFrame, y_train: pd.Series) -> LinearModelWrapper:
        self.model.fit(X_train[self._exog_cols], y_train)
        return self

    def predict(self, X_test: pd.DataFrame) -> ndarray:
        return self.model.predict(X_test[self._exog_cols])


class LinearModelWrapperWithIndexName(LinearModelWrapper):
    def __init__(self, index_name: str, **kwargs):
        super().__init__(**kwargs)
        self._index_name = index_name

    def predict(self, X_test: pd.DataFrame) -> pd.Series:
        return pd.Series(
            self.model.predict(X_test[self._exog_cols]),
            index=pd.Index(X_test.index, name=self._index_name),
        )


class ClassifierModelWrapper(ModelBase):
    def __init__(self, exog_cols: List[str]) -> None:
        self._exog_cols = exog_cols
        self.model = SVC()

    def fit(self, X_train: pd.DataFrame, y_train: pd.Series) -> ClassifierModelWrapper:
        self.model.fit(X_train[self._exog_cols], y_train)
        return self

    def predict(self, X_test: pd.DataFrame) -> ndarray:
        return self.model.predict(X_test[self._exog_cols])
